# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import namedtuple
from itertools import islice, repeat, chain

from javalang.tree import Node
from typing import Union, Any, Set, List, Iterator, Tuple, Dict, cast
from networkx import DiGraph, dfs_labeled_edges, dfs_preorder_nodes  # type: ignore
from deprecated import deprecated  # type: ignore

from aibolit.ast_framework.ast_node_type import ASTNodeType
from aibolit.ast_framework._auxiliary_data import javalang_to_ast_node_type, attributes_by_node_type, ASTNodeReference
from aibolit.ast_framework.ast_node import ASTNode

MethodInvocationParams = namedtuple('MethodInvocationParams', ['object_name', 'method_name'])

MemberReferenceParams = namedtuple('MemberReferenceParams', ('object_name', 'member_name', 'unary_operator'))

BinaryOperationParams = namedtuple('BinaryOperationParams', ('operation', 'left_side', 'right_side'))


class AST:
    def __init__(self, networkx_tree: DiGraph, root: int):
        self.tree = networkx_tree
        self.root = root

    @staticmethod
    def build_from_javalang(javalang_ast_root: Node) -> 'AST':
        tree = DiGraph()
        javalang_node_to_index_map: Dict[Node, int] = {}
        root = AST._add_subtree_from_javalang_node(tree, javalang_ast_root,
                                                   javalang_node_to_index_map)
        AST._change_node_attributes_to_index(tree, javalang_node_to_index_map)
        return AST(tree, root)

    def __str__(self) -> str:
        printed_graph = ''
        depth = 0
        print_step = 4
        for _, destination, edge_type in dfs_labeled_edges(self.tree, self.root):
            if edge_type == 'forward':
                if depth > 0:
                    printed_graph += ' ' * depth + '|---'
                node_type = self.get_type(destination)
                printed_graph += str(node_type)
                if node_type == ASTNodeType.STRING:
                    printed_graph += ': ' + self.get_attr(destination, 'string')
                printed_graph += '\n'
                depth += print_step
            elif edge_type == 'reverse':
                depth -= print_step
        return printed_graph

    def get_root(self) -> ASTNode:
        return ASTNode(self.tree, self.root)

    def __iter__(self) -> Iterator[ASTNode]:
        for node_index in self.tree.nodes:
            yield ASTNode(self.tree, node_index)

    def get_subtrees(self, root_type: ASTNodeType) -> Iterator['AST']:
        '''
        Yields subtrees with given type of the root.
        If such subtrees are one including the other, only the larger one is
        going to be in resulted sequence.
        '''
        is_inside_subtree = False
        current_subtree_root = -1  # all node indexes are positive
        subtree: List[int] = []
        for _, destination, edge_type in dfs_labeled_edges(self.tree, self.root):
            if edge_type == 'forward':
                if is_inside_subtree:
                    subtree.append(destination)
                elif self.tree.nodes[destination]['node_type'] == root_type:
                    subtree.append(destination)
                    is_inside_subtree = True
                    current_subtree_root = destination
            elif edge_type == 'reverse' and destination == current_subtree_root:
                is_inside_subtree = False
                yield AST(self.tree.subgraph(subtree), current_subtree_root)
                subtree = []
                current_subtree_root = -1

    def get_subtree(self, node: ASTNode) -> 'AST':
        subtree_nodes_indexes = dfs_preorder_nodes(self.tree, node.node_index)
        subtree = self.tree.subgraph(subtree_nodes_indexes)
        return AST(subtree, node.node_index)

    @deprecated(reason='Use ASTNode functionality instead.')
    def children_with_type(self, node: int, child_type: ASTNodeType) -> Iterator[int]:
        '''
        Yields children of node with given type.
        '''
        for child in self.tree.succ[node]:
            if self.tree.nodes[child]['node_type'] == child_type:
                yield child

    @deprecated(reason='Use ASTNode functionality instead.')
    def list_all_children_with_type(self, node: int, child_type: ASTNodeType) -> List[int]:
        list_node: List[int] = []
        for child in self.tree.succ[node]:
            list_node = list_node + self.list_all_children_with_type(child, child_type)
            if self.tree.nodes[child]['node_type'] == child_type:
                list_node.append(child)
        return sorted(list_node)

    @deprecated(reason='Use ASTNode functionality instead.')
    def all_children_with_type(self, node: int, child_type: ASTNodeType) -> Iterator[int]:
        '''
        Yields all children of node with given type.
        '''
        for child in self.list_all_children_with_type(node, child_type):
            yield child

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_first_n_children_with_type(self, node: int, child_type: ASTNodeType, quantity: int) -> List[int]:
        '''
        Returns first quantity of children of node with type child_type.
        Resulted list is padded with None to length quantity.
        '''
        children_with_type = (child for child in self.tree.succ[node] if self.get_type(child) == child_type)
        children_with_type_padded = chain(children_with_type, repeat(None))
        return list(islice(children_with_type_padded, 0, quantity))

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_binary_operation_name(self, node: int) -> str:
        assert(self.get_type(node) == ASTNodeType.BINARY_OPERATION)
        name_node, = islice(self.children_with_type(node, ASTNodeType.STRING), 1)
        return self.get_attr(name_node, 'string')

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_line_number_from_children(self, node: int) -> int:
        for child in self.tree.succ[node]:
            cur_line = self.get_attr(child, 'line')
            if cur_line is not None:
                return cur_line
        return 0

    @deprecated(reason='Use get_proxy_nodes instead.')
    def get_nodes(self, type: Union[ASTNodeType, None] = None) -> Iterator[int]:
        for node in self.tree.nodes:
            if type is None or self.tree.nodes[node]['node_type'] == type:
                yield node

    def get_proxy_nodes(self, *types: ASTNodeType) -> Iterator[ASTNode]:
        for node in self.tree.nodes:
            if len(types) == 0 or self.tree.nodes[node]['node_type'] in types:
                yield ASTNode(self.tree, node)

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_attr(self, node: int, attr_name: str, default_value: Any = None) -> Any:
        return self.tree.nodes[node].get(attr_name, default_value)

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_type(self, node: int) -> ASTNodeType:
        return self.get_attr(node, 'node_type')

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_method_invocation_params(self, invocation_node: int) -> MethodInvocationParams:
        assert(self.get_type(invocation_node) == ASTNodeType.METHOD_INVOCATION)
        # first two STRING nodes represent object and method names
        children = list(self.children_with_type(invocation_node, ASTNodeType.STRING))
        if len(children) == 1:
            return MethodInvocationParams('', self.get_attr(children[0], 'string'))

        return MethodInvocationParams(self.get_attr(children[0], 'string'),
                                      self.get_attr(children[1], 'string'))

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_member_reference_params(self, member_reference_node: int) -> MemberReferenceParams:
        assert(self.get_type(member_reference_node) == ASTNodeType.MEMBER_REFERENCE)
        params = [self.get_attr(child, 'string') for child in
                  self.children_with_type(member_reference_node, ASTNodeType.STRING)]

        member_reference_params: MemberReferenceParams
        if len(params) == 1:
            member_reference_params = MemberReferenceParams(object_name='', member_name=params[0],
                                                            unary_operator='')
        elif len(params) == 2:
            member_reference_params = MemberReferenceParams(object_name=params[0], member_name=params[1],
                                                            unary_operator='')
        elif len(params) == 3:
            member_reference_params = MemberReferenceParams(unary_operator=params[0], object_name=params[1],
                                                            member_name=params[2])
        else:
            raise ValueError('Node has 0 or more then 3 children with type "STRING": ' + str(params))

        return member_reference_params

    @deprecated(reason='Use ASTNode functionality instead.')
    def get_binary_operation_params(self, binary_operation_node: int) -> BinaryOperationParams:
        assert(self.get_type(binary_operation_node) == ASTNodeType.BINARY_OPERATION)
        operation_node, left_side_node, right_side_node = self.tree.succ[binary_operation_node]
        return BinaryOperationParams(self.get_attr(operation_node, 'string'), left_side_node, right_side_node)

    @staticmethod
    def _add_subtree_from_javalang_node(tree: DiGraph, javalang_node: Union[Node, Set[Any], str],
                                        javalang_node_to_index_map: Dict[Node, int]) -> int:
        node_index, node_type = AST._add_javalang_node(tree, javalang_node)
        if node_index != AST._UNKNOWN_NODE_TYPE and \
           node_type not in {ASTNodeType.COLLECTION, ASTNodeType.STRING}:
            javalang_standard_node = cast(Node, javalang_node)
            javalang_node_to_index_map[javalang_standard_node] = node_index
            AST._add_javalang_children(tree, javalang_standard_node.children, node_index,
                                       javalang_node_to_index_map)
        return node_index

    @staticmethod
    def _add_javalang_children(tree: DiGraph, children: List[Any], parent_index: int,
                               javalang_node_to_index_map: Dict[Node, int]) -> None:
        for child in children:
            if isinstance(child, list):
                AST._add_javalang_children(tree, child, parent_index, javalang_node_to_index_map)
            else:
                child_index = AST._add_subtree_from_javalang_node(tree, child, javalang_node_to_index_map)
                if child_index != AST._UNKNOWN_NODE_TYPE:
                    tree.add_edge(parent_index, child_index)

    @staticmethod
    def _add_javalang_node(tree: DiGraph, javalang_node: Union[Node, Set[Any], str]) -> Tuple[int, ASTNodeType]:
        node_index = AST._UNKNOWN_NODE_TYPE
        node_type = ASTNodeType.UNKNOWN
        if isinstance(javalang_node, Node):
            node_index, node_type = AST._add_javalang_standard_node(tree, javalang_node)
        elif isinstance(javalang_node, set):
            node_index = AST._add_javalang_collection_node(tree, javalang_node)
            node_type = ASTNodeType.COLLECTION
        elif isinstance(javalang_node, str):
            node_index = AST._add_javalang_string_node(tree, javalang_node)
            node_type = ASTNodeType.STRING

        return node_index, node_type

    @staticmethod
    def _add_javalang_standard_node(tree: DiGraph, javalang_node: Node) -> Tuple[int, ASTNodeType]:
        node_index = len(tree) + 1
        node_type = javalang_to_ast_node_type[type(javalang_node)]

        attr_names = attributes_by_node_type[node_type]
        attributes = {attr_name: getattr(javalang_node, attr_name) for attr_name in attr_names}

        attributes['node_type'] = node_type
        attributes['line'] = javalang_node.position.line if javalang_node.position is not None else None

        tree.add_node(node_index, **attributes)
        return node_index, node_type

    @staticmethod
    def _add_javalang_collection_node(tree: DiGraph, collection_node: Set[Any]) -> int:
        node_index = len(tree) + 1
        tree.add_node(node_index, node_type=ASTNodeType.COLLECTION)
        # we expect only strings in collection
        # we add them here as children
        for item in collection_node:
            if type(item) == str:
                string_node_index = AST._add_javalang_string_node(tree, item)
                tree.add_edge(node_index, string_node_index)
            elif item is not None:
                raise ValueError('Unexpected javalang AST node type {} inside \
                                 "COLLECTION" node'.format(type(item)))
        return node_index

    @staticmethod
    def _add_javalang_string_node(tree: DiGraph, string_node: str) -> int:
        node_index = len(tree) + 1
        tree.add_node(node_index, node_type=ASTNodeType.STRING, string=string_node)
        return node_index

    @staticmethod
    def _change_node_attributes_to_index(tree: DiGraph,
                                         javalang_node_to_index_map: Dict[Node, int]) -> None:
        '''
        All javalang fields are stored as is. If some of those fields were pointing to Node
        we change that attribute to have a networkx node index, which represent that node.
        '''
        for node, attributes in tree.nodes.items():
            for attribute_name in attributes:
                attribute_value = attributes[attribute_name]
                if isinstance(attribute_value, Node):
                    node_reference = ASTNodeReference(javalang_node_to_index_map[attribute_value])
                    tree.add_node(node, **{attribute_name: node_reference})
                elif isinstance(attribute_value, list) and \
                        all(isinstance(item, Node) for item in attribute_value):
                    node_references = [ASTNodeReference(javalang_node_to_index_map[item])
                                       for item in attribute_value]
                    tree.add_node(node, **{attribute_name: node_references})

    _UNKNOWN_NODE_TYPE = -1





class MutableIndex:

    def __init__(self):
        pass

    def value(self, filename: str):
        """
        Traverse over AST tree and finds loops with mutable index
        :return:
        List of line number of loops with mutable index
        @todo #147:30min Implement MutableIndex pattern
         We need to implement the code on mutable index pattern. It must return the number
         of the lines where we have an index of a for loop being mutated. After implementing,
         the method, enable all tests in test_mutable_index.py
        """
        return []



from cchardet import detect  # type: ignore


def detect_encoding_of_file(filename: str):
    with open(filename, 'rb') as target_file:
        return detect_encoding_of_data(target_file.read())


def detect_encoding_of_data(data: bytes):
    return detect(data)['encoding']


def read_text_with_autodetected_encoding(filename: str):
    with open(filename, 'rb') as target_file:
        data = target_file.read()

    if not data:
        return ''  # In case of empty file, return empty string

    encoding = detect_encoding_of_data(data) or 'utf-8'
    return data.decode(encoding)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the 'Software'), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED 'AS IS', WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase, skip
from pathlib import Path

from aibolit.ast_framework.java_package import JavaPackage


@skip('JavaClassMethod is deprecated')
class JavaClassMethodTestCase(TestCase):
    def test_method_name(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'TwoClasses.java')
        java_class = java_package.java_classes['Second']
        self.assertEqual(java_class.methods.keys(), {'Decrement', 'Increment'})

    def test_used_methods(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'MethodUseOtherMethodExample.java')
        java_class = java_package.java_classes['MethodUseOtherMethod']
        for method_name, methods in java_class.methods.items():
            with self.subTest(f'Method: {method_name}'):
                self.assertEqual(len(methods), 1)
                method = next(iter(methods))
                self.assertEqual(JavaClassMethodTestCase._used_methods_by_name[method_name],
                                 method.used_methods.keys())

    def test_method_parameters(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'MethodUseOtherMethodExample.java')
        java_class = java_package.java_classes['MethodUseOtherMethod']
        for method_name, methods in java_class.methods.items():
            with self.subTest(f'Method: {method_name}'):
                self.assertEqual(len(methods), 1)
                method = next(iter(methods))
                self.assertEqual(JavaClassMethodTestCase._methods_parameters[method_name],
                                 method.parameters.keys())

    def test_used_fields(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'MethodUseOtherMethodExample.java')
        java_class = java_package.java_classes['MethodUseOtherMethod']
        for method_name, methods in java_class.methods.items():
            with self.subTest(f'Method: {method_name}'):
                self.assertEqual(len(methods), 1)
                method = next(iter(methods))
                self.assertEqual(JavaClassMethodTestCase._used_fields_by_name[method_name],
                                 method.used_fields.keys())

    _used_methods_by_name = {
        'useOnlyMethods1': {'useOnlyMethods2'},
        'useOnlyMethods2': {'useOnlyMethods1'},
        'getField': set(),
        'setField': set(),
        'standAloneMethod': set(),
        'shadowing': set(),
    }

    _methods_parameters = {
        'useOnlyMethods1': {'x'},
        'useOnlyMethods2': {'x'},
        'getField': set(),
        'setField': {'value'},
        'standAloneMethod': set(),
        'shadowing': {'redundantField'},
    }

    _used_fields_by_name = {
        'useOnlyMethods1': set(),
        'useOnlyMethods2': set(),
        'getField': {'connectingField'},
        'setField': {'connectingField'},
        'standAloneMethod': set(),
        'shadowing': set(),
    }



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import itertools
from collections import defaultdict
from collections import namedtuple

import javalang

from aibolit.utils.java_parser import JavalangImproved

ExceptionInfo = namedtuple('ExceptionInfo', 'func_name, catch_list, throws_list, line_number')


class RedundantCatch:
    """
    Find pattern when a method in Java class throws an exception,
    and the exception of the same type is caught inside the method.

    E.g.,
    class Book {
    void foo() throws IOException {
        try {
          Files.readAllBytes();
        } catch (IOException e)
          { // here
          // do something
          }
        }
    }
    Here, the method foo() throws IOException, but we catch it inside the method
    """
    def __init__(self):
        pass

    def value(self, filename):
        """
        Find the mentioned-above pattern
        :param filename: filename of Java file
        :return: code lines of try statement where it was found
        """
        total_code_lines = set()
        obj = JavalangImproved(filename)
        items = obj.tree_to_nodes()
        try_nodes = defaultdict(list)
        method_nodes = {}
        for x in items:
            # Line break occurred before a binary operator (W503)
            # But this rule goes against the PEP 8 recommended style, so
            # replace isinstanceof with variable
            is_instance_meth_decl = isinstance(x.node, javalang.tree.MethodDeclaration)
            is_instance_try_stat = isinstance(x.node, javalang.tree.TryStatement)
            is_instance_ctor_decl = isinstance(x.node, javalang.tree.ConstructorDeclaration)
            is_instance_lambda = isinstance(x.node, javalang.tree.LambdaExpression)
            if is_instance_try_stat and x.method_line and not is_instance_lambda:
                # If we do not have a line for method, we ignore this method
                try_nodes[x.method_line].append(x)
            elif (is_instance_meth_decl or is_instance_ctor_decl) and x.method_line and not is_instance_lambda:
                # If we do not have a line for method, we ignore this method
                method_nodes[x.method_line] = x

        for method_line, iter_nodes in sorted(try_nodes.items(), key=lambda x: x[1][0].line):
            for try_node in iter_nodes:
                method_node = method_nodes.get(method_line)

                if not method_node or not method_node.node.throws:
                    continue

                catch_list = []
                ei = ExceptionInfo(
                    func_name=method_node.node.name,
                    catch_list=catch_list,
                    throws_list=method_node.node.throws,
                    line_number=method_node.node.position.line
                )
                if try_node.node.catches:
                    catch_classes = [x.parameter.types for x in try_node.node.catches]
                    classes_exception_list = list(itertools.chain(*catch_classes))
                    ei.catch_list.extend(classes_exception_list)

                    lines_number = set([
                        try_node.line for c in ei.catch_list if c in ei.throws_list
                    ])
                    total_code_lines.update(lines_number)

        return sorted(total_code_lines)



import os
import unittest
from aibolit.patterns.method_siblings.method_siblings import MethodSiblings
from pathlib import Path


class TestMethodSiblingsSiblings(unittest.TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_simple_method_siblings(self):
        self.assertEqual(
            [2, 5], MethodSiblings().value(Path(self.dir_path, 'SimpleMethodSiblings.java'))
        )

    def test_find_alternate_method_siblings(self):
        self.assertEqual(
            [2, 8], MethodSiblings().value(Path(self.dir_path, 'AlternateMethodSiblings.java'))
        )



from typing import List, Tuple, Any, Union, TypeVar, Type
from javalang.tree import ClassDeclaration, InterfaceDeclaration, MethodDeclaration, \
    MemberReference, FieldDeclaration, MethodInvocation, This, Node, LocalVariableDeclaration

FldExh = Tuple[str, Tuple[str, str]]
MthExh = Tuple[str, Tuple[Tuple[str, str], ...]]
HasMember = Union[MemberReference, MethodInvocation]
HasSelector = Union[MemberReference, MethodInvocation, This]
T = TypeVar('T', bound=Node)
S = TypeVar('S', HasSelector, HasMember)
MthNodes = Tuple[tuple, MethodDeclaration]
Nodes = Tuple[tuple, T]
AnyField = Union[FieldDeclaration, LocalVariableDeclaration]


class Filters:

    def __init__(self):
        pass

    def filter_node_lvl(self, node: Node, javalang_class: Type[Node]) -> List[Nodes]:
        """Filters nodes by desired javalang class.

        Gets node(node) of any javalang.Tree type and filters it by
        desired type(javalang_class).
        Returns a generator with (path, node) inside.
        """
        temp_list = []
        for filtered_path, filtered_node in node.filter(javalang_class):
            if self.get_class_depth(filtered_path) == 1:
                temp_list.append((filtered_path, filtered_node))
        return temp_list

    @staticmethod
    def filter_getters_setters(method_node_list: List[MthNodes]) -> List[MthNodes]:
        """Filters nodes by name.

        Gets list of nodes of "MethodDeclaration" type and filters it by
        name, so that no methods with name starting with "set" or "get"
        go to return list.
        Returns a generator with (path, node) inside.
        """

        # ToDo: implement get/set detection with .body
        temp_list = []
        for path, node in method_node_list:
            if node.name.startswith(('get', 'set')):
                pass
            else:
                temp_list.append((path, node))
        return temp_list

    @staticmethod
    def get_class_depth(path: tuple) -> int:
        """Returns an int displaying level of given node's nesting level.

        Gets a node of any javalang.tree type and calculates it's nesting level
        Returns an int.
        """

        class_level = 0
        for step in path:
            if isinstance(step, (ClassDeclaration, InterfaceDeclaration, MethodDeclaration)):
                class_level += 1
        return class_level

    @staticmethod
    def exhaust_method(method_node: MethodDeclaration) -> MthExh:
        """ Exhausts name and input vars, types for given MethodDeclaration node.

        Returns a tuple containing name and all parameters
        that given method gets as an input.
        """

        parameter_list = []
        name: str = method_node.name
        for parameter in method_node.parameters:
            parameter_list.append((parameter.name, parameter.type.name))
        parameter_tuple: Tuple[Tuple[str, str], ...] = tuple(parameter_list)
        return name, parameter_tuple

    @staticmethod
    def exhaust_field(field_node: AnyField) -> FldExh:
        """ Exhausts name and type for given FieldDeclaration or LocalVariableDeclaration node.

        Returns a tuple containing name and type of field.
        """

        # ToDo: get rid of'type' in parameter_tuple
        name = field_node.declarators[0].name
        try:
            parameter_tuple: Tuple[str, str] = ('type', field_node.type.name)
        except AttributeError:
            return "", ("", "")
        return name, parameter_tuple

    @staticmethod
    def get_arguments(invocation_node: MethodInvocation) -> Tuple[List[str], List[str]]:
        """Gets arguments passed to given MethodInvocation node.

        Returns two tuples containing all arguments and methods passed
        to given MethodInvocation node.
        """

        list_of_funcs = []
        list_of_fields = []
        for argument in invocation_node.arguments:  # type: ignore
            if isinstance(argument, MethodInvocation):
                list_of_funcs.append(argument.member)
            elif isinstance(argument, MemberReference):
                list_of_fields.append(argument.member)
        return list_of_funcs, list_of_fields

    @staticmethod
    def clean_for_repetitions(list_of_exhaust: List[Any]) -> List[Any]:
        """Gets any list and removes all repetitions.

        Returns list with no repetitive objects.
        """

        list_of_exhaust = list(dict.fromkeys(list_of_exhaust))
        return list_of_exhaust



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase
from pathlib import Path

from aibolit.ast_framework.java_package import JavaPackage


class CFGBuilderTestCase(TestCase):

    def test_cfg_of_method(self):
        java_package = JavaPackage(Path(__file__).parent.absolute() / 'SimpleClass.java')
        java_class = java_package.java_classes['Simple']
        methods = java_class.methods['Increment']
        self.assertEqual(len(methods), 1)
        method = next(iter(methods))
        self.assertEqual(method.cfg.size(), 2)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
from aibolit.ast_framework import ASTNodeType, AST
from aibolit.utils.ast_builder import build_ast
from typing import List
from aibolit.ast_framework.ast_node import ASTNode


class EmptyRethrow:
    '''
    Check if we throw the same exception as it was caught
    '''
    def _process_catch(self, ast: AST, catch_clauses: List[ASTNode]):
        lines: List[int] = []
        for catch_clause in catch_clauses:
            throw_statements = ast.get_subtree(catch_clause).get_proxy_nodes(ASTNodeType.THROW_STATEMENT)
            for throw_stat in throw_statements:
                if throw_stat.expression.node_type == ASTNodeType.MEMBER_REFERENCE \
                   and throw_stat.expression.member == catch_clause.parameter.name:
                    lines.append(throw_stat.line)
        return lines

    def value(self, filename) -> List[int]:
        total_code_lines: List[int] = []
        ast = AST.build_from_javalang(build_ast(filename))
        for try_node in ast.get_proxy_nodes(ASTNodeType.TRY_STATEMENT):
            catch_clauses = try_node.catches
            if catch_clauses:
                total_code_lines.extend(self._process_catch(ast, catch_clauses))

        return sorted(total_code_lines)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import unittest
from aibolit.metrics.local_methods_calls.local_methods_calls import LocalMethodsCalls
from pathlib import Path
from unittest import TestCase


class LocalMethodsCallsTest(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_no_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'NoLocalMethodCalls.java')),
            0,
            'Could not calculate how many local methods are called when none'
        )

    @unittest.skip('Not implemented')
    def test_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'LocalMethodCalls.java')),
            4,
            'Could not calculate how many local method are called when they exist'
        )

    @unittest.skip('Not implemented')
    def test_inner_class_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'InnerClassLocalMethodCalls.java')),
            2,
            'Could not calculate how many local methods called when they exist in inner class'
        )

    @unittest.skip('Not implemented')
    def test_double_local_method_calls(self):
        self.assertEqual(
            LocalMethodsCalls().value(Path(self.dir_path, 'DoubleLocalMethodCalls.java')),
            2,
            'Could not calculate how many local methods called when they are called more than once'
        )



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.


class ExternalMethodsCalled:
    """
        Measure the number of external methods called by the class.
        @todo #183:30min Implement external method called metric
         In the methods in the class an external method for this class can
         be called. So number of external methods called by all the methods
         in the class is suggested as a metric. Implement this metric and then
         enable tests in test_external_methods_called.py
    """

    def __init__(self):
        pass

    def value(self, filename: str):
        return 0



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from pathlib import Path
from aibolit.metrics.lcom4.lcom4 import LCOM4


class TestLCOM4(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    pattern = LCOM4()

    def test_class_with_empty_method(self):
        name = 'EmptyMethod.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_constructor(self):
        name = 'Constructor.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)

    def test_scope_with_anonymous(self):
        name = 'ScopeAnonymous.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_simple(self):
        name = 'Simple.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)

    def test_getter_setter(self):
        name = 'GetterSetter.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        # We should ignore all setters and getters
        self.assertEqual(lcom4_val, 1)

    def test_scope(self):
        name = 'Scope.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_overloaded(self):
        name = 'Overloaded.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)

    def test_class_with_chain(self):
        name = 'MethodChain.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 2)

    def test_overloaded_diff(self):
        name = 'OverloadedDiffComp.java'
        lcom4_val = self.pattern.value(Path(self.dir_path, name))
        self.assertEqual(lcom4_val, 1)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import javalang

from aibolit.utils.ast_builder import build_ast


class ThisFinder:

    def __expr_stat(self, expr, flag_this, flag_else):
        '''function to work with StatementExpression block'''
        if isinstance(expr.expression, javalang.tree.ExplicitConstructorInvocation):
            if flag_this + flag_else > 0:
                return 1, flag_this, flag_else
            flag_this = 1
        else:
            if flag_this > 0:
                return 1, flag_this, flag_else
        flag_else = 1
        return 0, flag_this, flag_else

    def __try_stat(self, expr, flag_this, flag_else):
        '''function to work with TryStatement block'''
        if (expr.resources is not None) or \
            (expr.catches is not None and expr.catches[0].block != []) or \
                (expr.finally_block is not None):
            flag_else = 1
        try_exprs = expr.block
        for expr1 in try_exprs:
            if isinstance(expr1, javalang.tree.StatementExpression):
                res, flag_this, flag_else = self.__expr_stat(expr1, flag_this, flag_else)
                if res > 0:
                    return 1, flag_this, flag_else
            else:
                if flag_this > 0:
                    return 1, flag_this, flag_else
                flag_else = 1
        flag_else = 1
        return 0, flag_this, flag_else

    def __if_stat(self, expr, flag_this, flag_else):
        '''function to work with IfStatement block'''
        if expr.then_statement is not None:
            if hasattr(expr.then_statement, 'statements'):
                stmts = expr.then_statement.statements
            else:
                stmts = []
            res, flag_this, flag_else = self.__work_with_stats(stmts, flag_this, flag_else)
            if res > 0:
                return 1, flag_this, flag_else
        if expr.else_statement is not None:
            if isinstance(expr.else_statement, javalang.tree.IfStatement):
                res, flag_this, flag_else = self.__if_stat(expr.else_statement, flag_this, flag_else)
                if res > 0:
                    return 1, flag_this, flag_else
                return 0, flag_this, flag_else
            block = expr.else_statement
            res, flag_this, flag_else = self.__work_with_stats(block, flag_this, flag_else)
            if res > 0:
                return 1, flag_this, flag_else
        return 0, flag_this, flag_else

    # flake8: noqa
    def __work_with_stats(self, stats, flag_this, flag_else):
        '''function to work with objects in constructor'''
        for expr in stats:
            res = 0
            old_else = flag_else
            flag_else = 1
            if isinstance(expr, javalang.tree.TryStatement):
                res, flag_this, flag_else = self.__try_stat(expr, flag_this, old_else)
            elif isinstance(expr, javalang.tree.StatementExpression):
                res, flag_this, flag_else = self.__expr_stat(expr, flag_this, old_else)
            elif isinstance(expr, javalang.tree.IfStatement):
                res, flag_this, flag_else = self.__if_stat(expr, flag_this, flag_else)
            elif isinstance(expr, javalang.tree.ForStatement):
                if hasattr(expr.body, 'statements'):
                    res, flag_this, flag_else = self.__work_with_stats(expr.body.statements, flag_this, flag_else)
            elif isinstance(expr, javalang.tree.WhileStatement):
                if hasattr(expr.body, 'statements'):
                    res, flag_this, flag_else = self.__work_with_stats(expr.body.statements, flag_this, flag_else)
            elif isinstance(expr, javalang.tree.DoStatement):
                if hasattr(expr.body, 'statements'):
                    res, flag_this, flag_else = self.__work_with_stats(expr.body.statements, flag_this, flag_else)
            else:
                res = flag_this
            if res > 0:
                return 1, flag_this, flag_else
        return 0, flag_this, flag_else

    def value(self, filename: str):
        '''main function'''
        tree = build_ast(filename)
        num_str = []
        for _, node in tree.filter(javalang.tree.ConstructorDeclaration):
            number = node.position.line
            stats = node.children[-1]
            result, _, _ = self.__work_with_stats(stats, 0, 0)
            if result == 1:
                num_str.append(number)
        return sorted(list(set(num_str)))




import os
from unittest import TestCase
from aibolit.patterns.implements_multi.implements_multi import ImplementsMultiFinder
from pathlib import Path


class TestImplementsMulti(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    multi_finder = ImplementsMultiFinder()

    def test_one_class_with_types(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'AnimatableSplitDimensionPathValue.java'))
        self.assertEqual(lines, [])

    def test_two_classes(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'AnimatableTransform.java'))
        self.assertEqual(lines, [12])

    def test_implements_in_string(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'AuditEventModelProcessor.java'))
        self.assertEqual(lines, [])

    def test_implements_with_parantheses(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'BaseKeyframeAnimation.java'))
        self.assertEqual(lines, [])

    def test_implements_with_nested_parantheses(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'Configuration.java'))
        self.assertEqual(lines, [225])

    def test_implements_multi_classes(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'FillContent.java'))
        self.assertEqual(lines, [29])

    def test_implements_with_parantheses_multi(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'FJIterateTest.java'))
        self.assertEqual(lines, [601])

    def test_implements_with_parantheses_before(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'FJListProcedureRunner.java'))
        self.assertEqual(lines, [])

    def test_implements_in_comments(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'KeyProviderCryptoExtension.java'))
        self.assertEqual(lines, [])

    def test_implements_multi(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'OsSecureRandom.java'))
        self.assertEqual(lines, [42])

    def test_implements_three(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'RectangleContent.java'))
        self.assertEqual(lines, [22])

    def test_implements_many(self):
        lines = self.multi_finder.value(Path(self.dir_path, 'SequenceFile.java'))
        self.assertEqual(lines, [837])



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.metrics.spaces.SpaceCounter import IndentationCounter
from pathlib import Path


class TestSpaces(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    max_right = IndentationCounter(max_right=True)
    max_left = IndentationCounter(max_left=True)
    left_var = IndentationCounter(left_var=True)
    right_var = IndentationCounter(right_var=True)

    def test_class_with_best_ident(self):
        val = self.max_left.value(Path(self.dir_path, 'BestIdent.java'))
        self.assertEqual(val, 44)
        val = self.max_right.value(Path(self.dir_path, 'BestIdent.java'))
        self.assertEqual(val, 113)

    def test_class_without_left_spaces(self):
        val = self.max_left.value(Path(self.dir_path, 'NoLeftSpaces.java'))
        self.assertEqual(val, 0)
        val = self.max_right.value(Path(self.dir_path, 'NoLeftSpaces.java'))
        self.assertEqual(val, 55)

    def test_class_without_right_spaces(self):
        val = self.max_left.value(Path(self.dir_path, 'NoRightSpaces.java'))
        self.assertEqual(val, 57)
        val = self.max_right.value(Path(self.dir_path, 'NoRightSpaces.java'))
        self.assertEqual(val, 0)

    def test_class_with_equal_spaces_number(self):
        val = self.max_left.value(Path(self.dir_path, 'SameMean.java'))
        self.assertEqual(val, 4)
        val = self.max_right.value(Path(self.dir_path, 'SameMean.java'))
        self.assertEqual(val, 55)

    def test_class_with_tabs_and_spaces(self):
        val = self.max_left.value(Path(self.dir_path, 'SpacesAndTabs.java'))
        self.assertEqual(val, 8)
        val = self.max_right.value(Path(self.dir_path, 'SpacesAndTabs.java'))
        self.assertEqual(val, 59)

    def test_class_with_worst_ident(self):
        val = self.max_left.value(Path(self.dir_path, 'WorstIdentation.java'))
        self.assertEqual(val, 20)
        val = self.max_right.value(Path(self.dir_path, 'WorstIdentation.java'))
        self.assertEqual(val, 163)

    def test_empty_examples(self):
        val = self.max_left.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)
        val = self.max_right.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)
        val = self.left_var.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)
        val = self.right_var.value(Path(self.dir_path, 'ClusterDataSourceConfiguration.java'))
        self.assertEqual(val, 0)

    def test_one_point_in_series(self):
        val = self.max_left.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)
        val = self.max_right.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)
        val = self.left_var.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)
        val = self.right_var.value(Path(self.dir_path, 'package-info.java'))
        self.assertEqual(val, 0)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List, Tuple, Type

from javalang.ast import Node
from javalang.tree import CompilationUnit, BinaryOperation, Expression, Literal, ConstructorDeclaration

from aibolit.types_decl import LineNumber
from aibolit.utils.ast_builder import build_ast

Path = Tuple


_OP_EQUAL = "=="
_OP_NOT_EQUAL = "!="
_LT_NULL = "null"


class NullCheck(object):
    def value(self, filename: str) -> List[LineNumber]:
        tree: CompilationUnit = build_ast(filename)

        return self._traverse_node(tree)

    def _traverse_node(self, tree: CompilationUnit) -> List[LineNumber]:
        lines: List[LineNumber] = list()

        for path, node in tree.filter(BinaryOperation):
            if _is_null_check(node) and not _within_constructor(path):
                lines.append(node.operandr.position.line)

        return lines


def _is_null_check(node: BinaryOperation) -> bool:
    return node.operator in (_OP_EQUAL, _OP_NOT_EQUAL) and _is_null(node.operandr)


def _is_null(node: Expression) -> bool:
    return isinstance(node, Literal) and node.value == _LT_NULL


def _within_constructor(path: Path) -> bool:
    node_types: List[Type[Node]] = [type(p) for p in path[::2]]

    return ConstructorDeclaration in node_types



import os
import sys
from aibolit.config import Config
import tqdm

# TODO: fix all errors in the patterns/metrics and make these lists empty
EXCLUDE_PATTERNS = ['P31', 'P32']
EXCLUDE_METRICS = []

current_path: str = os.path.dirname(os.path.realpath(__file__))
print('Processed files in testing:')

for filename in tqdm.tqdm(os.listdir(current_path + '/samples')):

    for pattern in Config.get_patterns_config()['patterns']:
        if pattern['code'] in EXCLUDE_PATTERNS:
            continue
        try:
            path_to_file = os.path.join(current_path, 'samples', filename)
            pattern['make']().value(path_to_file)
        except Exception as e:
            print(
                f"Error in application of the pattern: {pattern['name']} \
                    with code {pattern['code']} for file {filename}",
                f'\nReason: {e}',
            )
            sys.exit(1)

    for metric in Config.get_patterns_config()['metrics']:
        if metric['code'] in EXCLUDE_METRICS:
            continue
        try:
            path_to_file = os.path.join(current_path, 'samples', filename)
            metric['make']().value(path_to_file)
        except Exception as e:
            print(
                f"Error in application of the metric: {metric['name']} \
                    with code {metric['code']} for file {filename}",
                f'\nReason: {e}',
            )
            sys.exit(1)

sys.exit(0)



from typing import List
from aibolit.types_decl import LineNumber


class NonFinalArgument:
    def __init__(self):
        pass

    def value(self, filename: str) -> List[LineNumber]:
        """
        Returns the line numbers of the file name which contains no final arguments
        :param filename: name of the file
        :return: number of the lines with non-final arguments
        @todo #146:30min Implement NonFinalArgument pattern.
         All method arguments must have final modifiers. Once we find an argument
         without final, it's a pattern. After implementing that, enable tests in
         test_non_final_argument.py
        """
        return []



import os
from unittest import TestCase
from aibolit.patterns.send_null.send_null import SendNull
from pathlib import Path


class TestSendNull(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    method_send_null_finder = SendNull()

    def test_one_send(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'BaseKeyframeAnimation.java'))
        self.assertEqual(lines, [146])

    def test_multi_level_invocation(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'Configuration.java'))
        self.assertEqual(lines, [379, 442, 549, 638, 656, 830, 866, 1362, 2393, 2988, 3080, 3492, 3855])

    def test_no_null_methods(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'FillContent.java'))
        self.assertEqual(lines, [])

    def test_simple_invocation(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'FJIterateTest.java'))
        self.assertEqual(lines, [493])

    def test_more_method_invocations(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'SequenceFile.java'))
        self.assertEqual(lines, [1186, 1201, 1217, 3285, 3298, 3367, 3537, 3550])

    def test_constructor_send_null(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'Constructor.java'))
        self.assertEqual(lines, [5, 14, 15, 16, 17, 18])

    def test_super_in_constructor_with_ternary_operator(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'AclPermissionParam.java'))
        self.assertEqual(lines, [46, 50])

    def test_this_with_ternary_operator(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'AddOp.java'))
        self.assertEqual(lines, [44, 48])

    def test_super_in_constructor_with_method_inv(self):
        lines = self.method_send_null_finder.value(Path(self.dir_path, 'ByteArrayMultipartFileEditor.java'))
        self.assertEqual(lines, [49])



from typing import Tuple, Dict, List
from typing import Tuple, Dict, List
import javalang

from aibolit.types_decl import LineNumber
from aibolit.utils.ast_builder import build_ast
from aibolit.utils.utils import RemoveComments


class StringConcatFinder:

    def __init__(self):
        pass

    # flake8: noqa: C901
    def value(self, filename: str) -> List[LineNumber]:

        lines = set()
        text = build_ast(filename)

        for _, node in text.filter(javalang.tree.BinaryOperation):
            if node.operator == '+':
                is_l_literal = isinstance(node.operandl, javalang.tree.Literal)
                is_r_literal = isinstance(node.operandr, javalang.tree.Literal)
                is_r_member = isinstance(node.operandr, javalang.tree.MemberReference)
                is_l_member = isinstance(node.operandl, javalang.tree.MemberReference)
                is_l_meth_inv = isinstance(node.operandl, javalang.tree.MethodInvocation)
                is_r_meth_inv = isinstance(node.operandr, javalang.tree.MethodInvocation)
                is_l_this = isinstance(node.operandl, javalang.tree.This)
                is_r_this = isinstance(node.operandr, javalang.tree.This)
                if is_l_literal and (is_r_member or is_r_meth_inv or is_r_this):
                    is_string_literal = '"' in node.operandl.value  # type: ignore
                    if is_string_literal:
                        if node.operandl.position:
                            lines.add(node.operandl.position.line)
                        elif node.operandr.position:
                            lines.add(node.operandr.position.line)
                        elif hasattr(node.operandl, '_position'):
                            lines.add(node.operandl._position.line)
                        elif hasattr(node.operandr, '_position'):
                            lines.add(node.operandr._position.line)
                elif is_r_literal and (is_l_member or is_l_meth_inv or is_l_this):
                    is_string_literal = '"' in node.operandr.value  # type: ignore
                    if is_string_literal:
                        if node.operandl.position:
                            lines.add(node.operandl.position.line)
                        elif node.operandr.position:
                            lines.add(node.operandr.position.line)
                        elif hasattr(node.operandr, '_position'):
                            lines.add(node.operandr._position.line)
                        elif hasattr(node.operandl, '_position'):
                            lines.add(node.operandl._position.line)

        return sorted(lines)



import os
import shutil
import subprocess
from pathlib import Path
from sklearn.model_selection import train_test_split  # type: ignore
import pickle
from random import randint
from aibolit.model.model import Dataset, TwoFoldRankingModel  # type: ignore
from aibolit.config import Config


def collect_dataset(args):
    """
    Run bash scripts to collect metrics and patterns for java files
    """

    def make_patterns(args, cur_work_dir):
        print('Compute patterns...')
        result = subprocess.run(['make', 'patterns'], stdout=subprocess.PIPE, encoding='utf-8', cwd=cur_work_dir)
        print(result.returncode)
        if result.returncode != 0:
            print(result.stderr)
            exit(3)
        else:
            print(result.stdout)
            if args.dataset_file:
                dataset_file_path = Path(cur_work_dir, args.dataset_file)
                if not dataset_file_path.parent.exists():
                    dataset_file_path.parent.mkdir(parents=True)
                shutil.copy(Path(Config.dataset_file()), dataset_file_path)
            else:
                dataset_file_path = Path(Config.dataset_file())
            print('dataset was saved to {}'.format(str(dataset_file_path.absolute())))

    def run_cmd(metrics_cmd, cur_work_dir):
        result = subprocess.run(metrics_cmd, stdout=subprocess.PIPE, encoding='utf-8', cwd=cur_work_dir)
        if result.returncode != 0:
            print(result.stderr)
            exit(1)
        else:
            print(result.stdout)

    # path to java files which will be analyzed
    java_folder = args.java_folder
    max_classes = args.max_classes
    os.chdir(Path(Config.home_aibolit_folder(), 'scripts'))
    if not java_folder:
        java_folder = Path('target/01').absolute()
        print('Analyzing {} dir:'.format(java_folder))
    cur_work_dir = Path(os.getcwd())
    print('Current working directory: ', cur_work_dir)
    print('Directory with JAVA classes: ', java_folder)
    print('Filtering java files...')

    filter_cmd = ['make', 'filter']
    metrics_cmd = ['make', 'metrics']
    merge_cmd = ['make', 'merge']
    build_halstead_cmd = ['make', 'build_halstead']
    make_hl_cmd = ['make', 'hl']

    if java_folder is not None:
        filter_cmd.append(f'dir={java_folder}')
        metrics_cmd.append(f'dir={java_folder}')
    if max_classes is not None:
        filter_cmd.append(f'max_classes={max_classes}')

    run_cmd(filter_cmd, cur_work_dir)
    print('Download PMD and compute metrics...')
    run_cmd(metrics_cmd, cur_work_dir)
    make_patterns(args, cur_work_dir)
    print('Building halstead.jar...')
    run_cmd(build_halstead_cmd, cur_work_dir)
    print('Calculating halstead metrics...')
    run_cmd(make_hl_cmd, cur_work_dir)

    print('Merge results and create dataset...')
    run_cmd(merge_cmd, cur_work_dir)


def train_process():
    """
    Define needed columns for dataset and run model training
    """
    config = Config.get_patterns_config()
    only_patterns = [
        x['code'] for x in list(config['patterns'])
        if x['code'] not in config['patterns_exclude']
    ]
    only_metrics = \
        [x['code'] for x in list(config['metrics'])
         if x['code'] not in config['metrics_exclude']] \
        + ['halstead volume']
    columns_features = only_metrics + only_patterns
    features_number = len(columns_features)
    print("Number of features: ", features_number)

    dataset = Dataset(only_patterns)
    dataset.preprocess_file()
    features_conf = {
        "features_order": dataset.feature_order,
        "patterns_only": only_patterns
    }

    X_train, X_test, y_train, y_test = train_test_split(dataset.input, dataset.target, test_size=0.3)
    model = TwoFoldRankingModel()
    model.fit(X_train, y_train)
    model.features_conf = features_conf

    save_model_file = Path(Config.folder_to_save_model_data(), 'model.pkl')
    print('Saving model to loaded model from file {}:'.format(save_model_file))
    with open(save_model_file, 'wb') as fid:
        pickle.dump(model, fid)

    load_model_file = Path(Config.folder_to_save_model_data(), 'model.pkl')
    print('Test loaded model from file {}:'.format(load_model_file))
    with open(load_model_file, 'rb') as fid:
        model_new = pickle.load(fid)
        print('Model has been loaded successfully')
        for x in X_test:
            snippet = [i for i in x] + [randint(1, 200)]
            preds, importances = model_new.informative(snippet)
            print(preds)
    path_with_logs = Path(os.getcwd(), 'catboost_info')
    print('Removing path with catboost logs {}'.format(path_with_logs))
    shutil.rmtree(path_with_logs)



from decimal import localcontext, ROUND_DOWN, Decimal
from typing import List

import numpy as np
import pandas as pd
from catboost import CatBoost
from sklearn.base import BaseEstimator
from sklearn.preprocessing import StandardScaler

from aibolit.config import Config


class Dataset:

    def __init__(self, only_patterns: List[str]):
        self.input = None
        self.target = None
        self.do_rename_columns = False
        self.only_patterns = only_patterns

    def preprocess_file(
            self,
            scale_ncss=True,
            scale=False,
            **kwargs):

        print('reading dataset from {}'.format(Config.dataset_file()))
        df = pd.read_csv(Config.dataset_file())
        df = df[~df["filename"].str.lower().str.contains("test")]
        config = Config.get_patterns_config()
        if self.do_rename_columns:
            p_codes = \
                [x['code'] for x in config['patterns']] \
                + ['lines' + x['code'] for x in config['patterns']]
            m_codes = [x['code'] for x in config['metrics']]
            keys = p_codes + m_codes
            vals = \
                [x['name'] for x in config['patterns']] \
                + ['lines' + x['name'] for x in config['patterns']] \
                + [x['name'] for x in config['metrics']]

            replace_dict = dict(zip(keys, vals))
            df = df.rename(replace_dict)
            df.columns = vals
            print('Columns renamed:' + df.head())

        df = df.dropna().drop_duplicates(subset=df.columns.difference(['filename']))
        df = df[(df.ncss > 20) & (df.ncss < 100) & (df.npath_method_avg < 100000.00)].copy().reset_index()

        df.drop('filename', axis=1, inplace=True)
        df.drop('index', axis=1, inplace=True)
        self.target = np.array(df[['M4']].values[:, 0], dtype=np.float64)
        if scale_ncss:
            new = pd.DataFrame(
                df[self.only_patterns].values / df['M2'].values.reshape((-1, 1)),
                columns=self.only_patterns
            )
            self.target /= df['M2'].values.reshape(-1)
        else:
            new = df[self.only_patterns].copy()
        if scale:
            self.input = pd.DataFrame(StandardScaler().fit_transform(new.values), columns=new.columns,
                                      index=new.index).values
        else:
            self.input = new.values

        self.feature_order = list(new.columns)


class TwoFoldRankingModel(BaseEstimator):

    def __init__(self):
        self.do_rename_columns = False
        self.model = None
        self.features_conf = None

    def fit(self, X, y, display=False):
        """
        Args:
            X: np.array with shape (number of snippets, number of patterns) or
                (number of patterns, ).
            y: np.array with shape (number of snippets,), array of snippets'
                complexity metric values
            display: bool, to output info about training or not
        """
        model = CatBoost()

        grid = {'learning_rate': [0.03, 0.1],
                'depth': [4, 6, 10],
                'l2_leaf_reg': [1, 3, 5, 7, 9]}

        model.grid_search(
            grid,
            X=X,
            y=y,
            verbose=display,
        )

        self.model = model
        self.model.fit(X, y.ravel(), logging_level='Silent')

    def sigmoid(self, x):
        return 1 / (1 + np.exp(-x))

    def __get_pairs(self, item, th: float, feature_importances=None):
        if not feature_importances:
            feature_importances = self.model.feature_importances_
        pattern_importances = item * feature_importances
        # mask discards not significant patterns
        th_mask = (self.sigmoid(pattern_importances) <= th) + 0
        pattern_importances = pattern_importances * th_mask
        order = np.arange(self.model.feature_importances_.size)
        return (pattern_importances, order)

    def __vstack_arrays(self, res):
        return np.vstack(res).T

    def calculate_score(self, X, quantity_func='log', th=1.0, feature_importances=None):
        """
        Args:
            X: np.array with shape ( number of patterns).
            quantity_func: str, type of function that will be applied to
                number of occurrences.
            th (float): Sensitivity of algorithm to recommend.
                0 - ignore all recommendations
                1 - use all recommendations

        Returns:
            ranked: np.array with shape (number of snippets, number of patterns)
                of sorted patterns in non-increasing order for each snippet of
                code.
        """

        X = X.copy()
        X = np.expand_dims(X, axis=0)

        ranked = []
        quantity_funcs = {
            'log': lambda x: np.log1p(x) / np.log(10),
            'exp': lambda x: np.exp(x + 1),
            'linear': lambda x: x,
        }

        try:
            item = quantity_funcs[quantity_func](X)
            pairs = self.__vstack_arrays(self.__get_pairs(item, th, feature_importances))
            pairs = pairs[pairs[:, 0].argsort()]
            ranked.append(pairs[:, 1].T.tolist()[::-1])
        except Exception:
            import traceback
            traceback.print_exc()
            raise Exception("Unknown func")

        return (np.array(ranked), pairs[:, 0].T.tolist()[::-1])

    def get_minimum(self, c1, c2, c3):
        """
        Args:
            c1, c2, c3: np.array with shape (number of snippets, ).
        Returns:
            c: np.array with shape (number of snippets, ) -
            elemental minimum of 3 arrays.
            number: np.array with shape (number of snippets, ) of
            arrays' numbers with minimum elements.            .
        """

        c = np.vstack((c1, c2, c3))

        return np.min(c, 0), np.argmin(c, 0)

    def informative(self, snippet, scale=True, th=1.0):
        """
        Args:
            snippet: np.array with shape (number of snippets, number of patterns + 1),
            because last column is ncss
        Returns:
            ranked: np.array with shape (number of snippets, number of patterns)
                of sorted patterns in non-increasing order for each snippet of
                code.
        """

        # remember it, since we will use `log` function for non-normalized input value
        patterns_orig = np.array(snippet[:-1])
        ncss = snippet[-1]

        if scale:
            snippet = patterns_orig / ncss
        else:
            snippet = patterns_orig

        k = snippet.size
        complexity = self.model.predict(snippet)
        importances = []
        for i in range(k):
            if snippet[i] == 0:
                # do not need to predict if we have 0
                importances.append((i, 0))
                continue
            temp_arr = snippet.copy()
            temp_arr[i] = temp_arr[i] - (1 / ncss)
            complexity_minus = self.model.predict(temp_arr)
            if complexity_minus < complexity:
                # complexity decreased
                with localcontext() as ctx:
                    ctx.rounding = ROUND_DOWN
                    abs_diff = abs(complexity - complexity_minus)
                    diff = Decimal(abs_diff).quantize(Decimal('0.001'))
                    diff = float(diff * 100)
            else:
                # complexity increased, we do not need such value, set to 0,
                # cause we need only patterns when complexity decreased
                diff = 0
            importances.append((i, diff))

        sorted_importances = dict(sorted(importances, key=lambda x: x[1], reverse=True))
        return sorted_importances.keys(), sorted_importances.values()




from typing import List

from javalang.tree import FieldDeclaration, ClassDeclaration

from aibolit.types_decl import LineNumber
from aibolit.utils.ast_builder import build_ast


class NonFinalAttribute:

    def __init__(self):
        pass

    def value(self, filename: str) -> List[LineNumber]:
        ret = []
        for _, clazz in build_ast(filename).filter(ClassDeclaration):
            for _, field in clazz.filter(FieldDeclaration):
                if 'final' not in field.modifiers:
                    ret.append(field.position.line)
        return list(dict.fromkeys(ret))



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.partial_synchronized.partial_synchronized import PartialSync


class PartialSynchronizedTest(TestCase):

    def test_simple(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Simple.java')
        self.assertEqual(lines, [22])

    def test_sync_with_array(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/ArrayIndex.java')
        self.assertEqual(lines, [89])

    def test_sync_with_method_chaining(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Chain.java')
        self.assertEqual(lines, [183])

    def test_sync_with_class_field(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/FieldClass.java')
        self.assertEqual(lines, [154])

    def test_sync_with_func_call(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/FuncCall.java')
        self.assertEqual(lines, [45])

    def test_sync_with_large_file(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Large.java')
        self.assertEqual(lines, [1359, 1546, 1584, 1606, 1628, 3347])

    def test_nested_sync(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/NestedSync.java')
        self.assertEqual(lines, [22, 24])

    def test_sync_with_class_this(self):
        """ Handles the case
        # synchronized(AsyncCallHandler.this)
        # {
        #     AsyncCallHandler.this.wait(waitTime);
        # }
        """
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Object.This.java')
        self.assertEqual(lines, [161])

    def test_sync_sequential(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SeveralCases.java')
        self.assertEqual(lines, [85, 91, 98])

    def test_sync_in_different_functions(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SeveralFunctions.java')
        self.assertEqual(lines, [22, 38, 40])

    def test_sync_with_statements_after_it(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SimpleThisAfter.java')
        self.assertEqual(lines, [22, 24])

    def test_sync_with_statements_before_it(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SimpleThisBefore.java')
        self.assertEqual(lines, [23, 25])

    def test_sync_with_empty_lines_after_it(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SimpleThisEmptyLineAfter.java')
        self.assertEqual(lines, [])

    def test_sync_with_empty_lines_before_it(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SimpleThisEmptyLineBefore.java')
        self.assertEqual(lines, [])

    def test_sync_with_static_class(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SyncStaticClass.java')
        self.assertEqual(lines, [52])

    def test_sync_with_nested_if_and_for(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/WIthIf.java')
        self.assertEqual(lines, [85])

    def test_sync_with_func_ref(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/SyncWithFunc.java')
        self.assertEqual(lines, [228])

    def test_ctor(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Constructor.java')
        self.assertEqual(lines, [412])

    def test_lambda(self):
        pattern = PartialSync()
        lines = pattern.value(os.path.dirname(os.path.realpath(__file__)) + '/Lambda.java')
        self.assertEqual(lines, [28])



import os
import unittest
from aibolit.patterns.var_siblings.var_siblings import VarSiblings
from pathlib import Path


class TestVarSiblings(unittest.TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_simple_var_siblings(self):
        self.assertEqual(
            [3, 4], VarSiblings().value(Path(self.dir_path, 'SimpleVarSiblings.java'))
        )

    def test_find_alternate_var_siblings(self):
        self.assertEqual(
            [3, 5], VarSiblings().value(Path(self.dir_path, 'AlternateVarSiblings.java'))
        )

    def test_find_length_4_var_siblings(self):
        self.assertEqual(
            [9, 10], VarSiblings().value(Path(self.dir_path, 'ShortVarSiblings.java'))
        )



from itertools import groupby
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.java_package import JavaPackage
from typing import List, Set
import re

only_increment_for: Set[ASTNodeType] = set([
    ASTNodeType.BREAK_STATEMENT,
    ASTNodeType.CONTINUE_STATEMENT,
    ASTNodeType.TERNARY_EXPRESSION,
    ASTNodeType.BINARY_OPERATION,
    ASTNodeType.METHOD_INVOCATION,
])

increment_and_nested_for: Set[ASTNodeType] = set([
    ASTNodeType.IF_STATEMENT,
    ASTNodeType.SWITCH_STATEMENT,
    ASTNodeType.FOR_STATEMENT,
    ASTNodeType.WHILE_STATEMENT,
    ASTNodeType.DO_STATEMENT,
    ASTNodeType.CATCH_CLAUSE,
])

logical_operators = ['&&', '||']


class CognitiveComplexity:

    def __init__(self):
        # store the name for the considered method declaration
        self.__method_name = None

    def _traverse_childs(self, ast: AST, node: int, nested_level: int) -> int:
        complexity = 0
        for each_child in ast.tree.succ[node]:
            complexity += self._get_complexity(ast, each_child, nested_level)
        return complexity

    def _check_if_statement(self, ast, expr, nested_level: int) -> int:
        '''function to work with IfStatement block'''
        complexity = 0
        all_childs = list([i for i in ast.tree.succ[expr]])
        complexity += self._get_complexity(ast, all_childs[0], 0)
        if len(all_childs) >= 2:
            complexity += nested_level + 1
            complexity += self._get_complexity(ast, all_childs[1], nested_level + 1)

        if len(all_childs) == 3:
            if ast.get_type(all_childs[2]) == ASTNodeType.IF_STATEMENT:
                complexity -= nested_level
                complexity += self._check_if_statement(ast, all_childs[2], nested_level)
            else:
                complexity += 1
                complexity += self._get_complexity(ast, all_childs[2], nested_level + 1)
        return complexity

    def _increment_logical_operators(self, ast: AST, binary_operation_node: int) -> int:
        complexity = 0
        logical_operators_sequence = self._create_logical_operators_sequence(ast, binary_operation_node)
        complexity += len(list(groupby(logical_operators_sequence)))
        return complexity

    def _create_logical_operators_sequence(self, ast: AST, binary_operation_node: int) -> List[str]:
        if ast.get_type(binary_operation_node) != ASTNodeType.BINARY_OPERATION:
            return []

        operator, left_side_node, right_side_node = ast.get_binary_operation_params(binary_operation_node)
        if operator not in logical_operators:
            return []

        left_sequence = self._create_logical_operators_sequence(ast, left_side_node)
        right_sequence = self._create_logical_operators_sequence(ast, right_side_node)
        return left_sequence + [operator] + right_sequence

    def _is_recursion_call(self, ast, node) -> bool:
        assert(ast.get_type(node) == ASTNodeType.METHOD_INVOCATION)
        if self.__method_name == self._get_node_name(ast, node):
            return True
        return False

    def _nested_methods(self, ast, node, nested_level: int) -> int:
        complexity = 0
        original_name = self.__method_name
        self.__method_name = self._get_node_name(ast, node)
        complexity += self._get_complexity(ast, node, nested_level + 1)
        self.__method_name = original_name
        return complexity

    def _process_not_nested_structure(self, ast: AST, each_block: int, nested_level: int) -> int:
        complexity = 0
        each_block_type = ast.get_type(each_block)
        if each_block_type == ASTNodeType.BINARY_OPERATION:
            bin_operator = ast.get_binary_operation_name(each_block)
            if bin_operator in logical_operators:
                complexity += self._increment_logical_operators(ast, each_block)

        elif each_block_type == ASTNodeType.METHOD_INVOCATION:
            is_recursion = self._is_recursion_call(ast, each_block)
            complexity += is_recursion

        else:
            complexity += 1
            complexity += self._traverse_childs(ast, each_block, nested_level)

        return complexity

    def _get_complexity(self, ast: AST, each_block: int, nested_level: int) -> int:
        each_block_name = self._get_node_name(ast, each_block)
        each_block_type = ast.get_type(each_block)
        complexity = 0

        if each_block_type == ASTNodeType.METHOD_DECLARATION and each_block_name != self.__method_name:
            complexity += self._nested_methods(ast, each_block, nested_level)

        elif each_block_type == ASTNodeType.IF_STATEMENT:
            complexity += self._check_if_statement(ast, each_block, nested_level)

        elif each_block_type in increment_and_nested_for:
            complexity += 1 + nested_level
            complexity += self._traverse_childs(ast, each_block, nested_level + 1)

        elif each_block_type in only_increment_for:
            complexity += self._process_not_nested_structure(ast, each_block, nested_level)

        else:
            complexity += self._traverse_childs(ast, each_block, nested_level)
        return complexity

    def _get_node_name(self, ast, node) -> str:
        extracted_name = None
        names = ast.children_with_type(node, ASTNodeType.STRING)
        for each_string in names:
            method_name = ast.get_attr(each_string, 'string')
            # Checking not to start with '/' is aimed to get
            # rid of comments, which are all childs of node.
            # We check the occurance any letter in name in order
            # to get rid of '' string and None.
            if not method_name.startswith('/') and re.search(r'[^\W\d]', method_name) is not None:
                extracted_name = method_name
                return extracted_name
        return ''

    def value(self, filename: str) -> int:
        complexity = 0
        p = JavaPackage(filename)
        for class_name in p.java_classes:
            tree = p.java_classes[class_name]
            for method_set in tree.methods.values():
                for method_ast in method_set:
                    self.__method_name = self._get_node_name(method_ast, method_ast.root)
                    complexity += self._get_complexity(method_ast, method_ast.root, 0)
        return complexity



import os
from unittest import TestCase
import unittest
from pathlib import Path
from aibolit.patterns.mutable_index.mutable_index import MutableIndex


class TestMutableIndex(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    @unittest.skip("Not implemented")
    def test_find_mutable_index_assignement(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexAssignment.java')),
            'Should match simple index assignent'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_prefix_increment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexPrefixIncrement.java')),
            'Should match index prefix increment (++i)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_prefix_decrement(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexPrefixDecrement.java')),
            'Should match index prefix decrement (--i)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_postfix_increment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexPostfixIncrement.java')),
            'Should match index postfix increment (i++)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_postfix_decrement(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexPostfixDecrement.java')),
            'Should match index postfix decrement (i--)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_addition_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexAdditionAssignment.java')),
            'Should match index addition assignment (+=)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_subtraction_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexSubtractionAssignment.java')),
            'Should match index subtraction assignment (-=)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_multiplication_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexMultiplicationAssignment.java')),
            'Should match index multiplication assignment (*=)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_division_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexDivisionAssignment.java')),
            'Should match index division assignment (/=)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_mod_assignment(self):
        self.assertEqual(
            [4],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexModAssignment.java')),
            'Should match index mod assignment (%=)'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_nested_for(self):
        self.assertEqual(
            [5],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexNestedFor.java')),
            'Should match mutable index in nested for'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_hidden_nested_for(self):
        self.assertEqual(
            [5],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexHiddenNestedFor.java')),
            'Should match mutable index hidden in nested for'
        )

    @unittest.skip("Not implemented")
    def test_find_mutable_index_no_block_for(self):
        self.assertEqual(
            [3],
            MutableIndex().value(Path(self.dir_path, 'MutableIndexNoBlockFor.java')),
            'Should match mutable index in for with no block'
        )



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from typing import List

from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNode, ASTNodeType


class ReturnNull:
    '''
    FInds all return statements that returns null directly or by ternary operator.
    NOTICE: nested ternary operators are not checked.
    '''

    def value(self, filename: str) -> List[int]:
        ast = AST.build_from_javalang(build_ast(filename))
        lines: List[int] = []
        for return_statement in ast.get_proxy_nodes(ASTNodeType.RETURN_STATEMENT):
            if self._check_null_return_statement(return_statement):
                lines.append(return_statement.line)

        return lines

    def _check_null_return_statement(self, return_statement: ASTNode) -> bool:
        # return statement with no expression `return;` does not return null
        if return_statement.expression is None:
            return False

        if return_statement.expression.node_type == ASTNodeType.TERNARY_EXPRESSION:
            return self._check_null_expression(return_statement.expression.if_true) or \
                self._check_null_expression(return_statement.expression.if_false)

        return self._check_null_expression(return_statement.expression)

    def _check_null_expression(self, expression: ASTNode) -> bool:
        return expression.node_type == ASTNodeType.LITERAL and expression.value == 'null'



import javalang

from aibolit.utils.ast_builder import build_ast


class PrivateStaticMethod:

    def __init__(self):
        pass

    def value(self, filename: str):
        return [
            node.position.line
            for _, node in build_ast(filename).filter(javalang.tree.MethodDeclaration)
            if all(elem in node.modifiers for elem in ['private', 'static'])
        ]




import os
from unittest import TestCase
from aibolit.patterns.public_static_method.public_static_method import PublicStaticMethod
from pathlib import Path


class TestNonFinalAttribute(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_non_final_atributes(self):
        lines = PublicStaticMethod().value(Path(self.dir_path, 'PublicStaticMethod.java'))
        self.assertEqual(1, len(lines))



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from networkx import DiGraph, disjoint_union  # type: ignore
from aibolit.ast_framework import AST, ASTNodeType
from typing import Tuple


NODE_TYPES = [
    ASTNodeType.ASSIGNMENT,
    ASTNodeType.RETURN_STATEMENT
]


def build_cfg(tree: AST) -> DiGraph:
    '''Create Control Flow Graph'''
    g = DiGraph()
    g.add_node(0)
    for node in tree:
        if node.node_type not in NODE_TYPES:
            continue
        _g = _mk_cfg_graph(node.node_type)
        g = _compose_two_graphs(g, _g)
    return g


def _mk_cfg_graph(node: ASTNodeType) -> Tuple[DiGraph, int]:
    '''Takes in Javalang statement and returns corresponding CFG'''
    g = DiGraph()
    g.add_node(0)
    return g


def _compose_two_graphs(g1: DiGraph, g2: DiGraph) -> DiGraph:
    '''Compose two graphs by creating the edge between last of the fist graph and fist of the second.
       We assume that node in the each graph G has order from 0 to len(G)-1
    '''
    g = disjoint_union(g1, g2)
    g.add_edge(len(g1) - 1, len(g1))
    return g



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase, skip
from pathlib import Path
from itertools import zip_longest

from aibolit.utils.ast_builder import build_ast
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.ast import MemberReferenceParams, MethodInvocationParams


class ASTTestSuite(TestCase):
    def test_parsing(self):
        ast = self._build_ast("SimpleClass.java")
        actual_node_types = [node.node_type for node in ast]
        self.assertEqual(actual_node_types,
                         ASTTestSuite._java_simple_class_preordered)

    def test_subtrees_selection(self):
        ast = self._build_ast("SimpleClass.java")
        subtrees = ast.get_subtrees(ASTNodeType.BASIC_TYPE)
        for actual_subtree, expected_subtree in \
                zip_longest(subtrees, ASTTestSuite._java_simple_class_basic_type_subtrees):
            with self.subTest():
                self.assertEqual([node.node_index for node in actual_subtree],
                                 expected_subtree)

    @skip('Method "get_member_reference_params" is deprecated')
    def test_member_reference_params(self):
        ast = self._build_ast("MemberReferencesExample.java")
        for node, expected_params in zip_longest(ast.get_nodes(ASTNodeType.MEMBER_REFERENCE),
                                                 ASTTestSuite._expected_member_reference_params):
            self.assertEqual(ast.get_member_reference_params(node), expected_params)

    @skip('Method "get_method_invocation_params" is deprecated')
    def test_method_invocation_params(self):
        ast = self._build_ast("MethodInvokeExample.java")
        for node, expected_params in zip_longest(ast.get_nodes(ASTNodeType.METHOD_INVOCATION),
                                                 ASTTestSuite._expected_method_invocation_params):
            self.assertEqual(ast.get_method_invocation_params(node), expected_params)

    def _build_ast(self, filename: str):
        javalang_ast = build_ast(str(Path(__file__).parent.absolute() / filename))
        return AST.build_from_javalang(javalang_ast)

    _java_simple_class_preordered = [
        ASTNodeType.COMPILATION_UNIT,
        ASTNodeType.CLASS_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.FIELD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.VARIABLE_DECLARATOR,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.METHOD_DECLARATION,
        ASTNodeType.COLLECTION,
        ASTNodeType.STRING,
        ASTNodeType.BASIC_TYPE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.STATEMENT_EXPRESSION,
        ASTNodeType.ASSIGNMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.LITERAL,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
        ASTNodeType.RETURN_STATEMENT,
        ASTNodeType.MEMBER_REFERENCE,
        ASTNodeType.STRING,
        ASTNodeType.STRING,
    ]

    _java_simple_class_basic_type_subtrees = [
        [8, 9],
        [17, 18],
    ]

    _expected_member_reference_params = [
        MemberReferenceParams('', 'block_variable', ''),
        MemberReferenceParams('', 'method_parameter', ''),
        MemberReferenceParams('', 'block_variable', '++'),
        MemberReferenceParams('', 'field', ''),
        MemberReferenceParams('', 'block_variable', ''),
        MemberReferenceParams('Something', 'outer_field', ''),
        MemberReferenceParams('', 'field', ''),
    ]

    _expected_method_invocation_params = [
        MethodInvocationParams(object_name='System.out', method_name='println'),
        MethodInvocationParams(object_name='', method_name='method1'),
    ]



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from cached_property import cached_property  # type: ignore
from typing import Dict, Set, TYPE_CHECKING
from networkx import DiGraph, dfs_tree  # type: ignore
from deprecated import deprecated  # type: ignore

from aibolit.utils.cfg_builder import build_cfg
from aibolit.ast_framework import AST, ASTNodeType
from aibolit.ast_framework.java_class_field import JavaClassField

if TYPE_CHECKING:
    from aibolit.ast_framework.java_class import JavaClass


@deprecated("This functionality must be transmitted to ASTNode")
class JavaClassMethod(AST):
    def __init__(self, tree: DiGraph, root: int, java_class: 'JavaClass'):
        self.tree = tree
        self.root = root
        self._java_class = java_class

    @cached_property
    def name(self) -> str:
        try:
            method_name = next(self.children_with_type(self.root, ASTNodeType.STRING))
            return self.tree.nodes[method_name]['string']
        except StopIteration:
            raise ValueError("Provided AST does not has 'STRING' node type right under the root")

    @property
    def java_class(self) -> 'JavaClass':
        return self._java_class

    @cached_property
    def parameters(self) -> Dict[str, AST]:
        parameters: Dict[str, AST] = {}
        for parameter_node in self.children_with_type(self.root, ASTNodeType.FORMAL_PARAMETER):
            parameter_name_node = next(iter(self.children_with_type(parameter_node, ASTNodeType.STRING)))
            parameter_name = self.get_attr(parameter_name_node, 'string')
            parameters[parameter_name] = AST(dfs_tree(self.tree, parameter_node), parameter_node)

        return parameters

    @cached_property
    def used_methods(self) -> Dict[str, Set['JavaClassMethod']]:
        method_invocation_nodes = self.get_nodes(ASTNodeType.METHOD_INVOCATION)
        used_method_invocation_params = (self.get_method_invocation_params(node) for node
                                         in method_invocation_nodes)
        used_local_method_invocation_params = (params for params in used_method_invocation_params
                                               if len(params.object_name) == 0)
        used_local_method_names = {params.method_name for params in used_local_method_invocation_params}

        return {method_name: self.java_class.methods[method_name] for method_name in self.java_class.methods
                if method_name in used_local_method_names}

    @cached_property
    def used_fields(self) -> Dict[str, JavaClassField]:
        used_member_reference_params = (self.get_member_reference_params(node) for node in
                                        self.get_nodes(ASTNodeType.MEMBER_REFERENCE))
        used_local_member_reference_names = {params.member_name for params in used_member_reference_params
                                             if params.object_name == ''}
        # Local member references may lead to method parameters instead of class fields if they have same names
        used_local_member_reference_names -= self.parameters.keys()
        class_fields = self.java_class.fields
        return {field_name: class_fields[field_name] for field_name in used_local_member_reference_names}

    @cached_property
    def cfg(self) -> DiGraph:
        '''Make Control Flow Graph representation of this method'''
        return build_cfg(self)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.nested_blocks.nested_blocks import NestedBlocks, BlockType
from pathlib import Path


class TestNestedBlocks(TestCase):
    depth_level = 2
    cur_file_dir = Path(os.path.realpath(__file__)).parent
    testClass = NestedBlocks(depth_level)

    def test_single_for_loop(self):
        file = str(Path(self.cur_file_dir, 'SingleFor.java'))
        self.assertEqual(self.testClass.value(file), [15, 19])

    def test_nested_for_loops(self):
        file = str(Path(self.cur_file_dir, 'NestedFor.java'))
        self.assertEqual(self.testClass.value(file), [22])

    def test_for_loops_in_different_methods(self):
        file = str(Path(self.cur_file_dir, 'DifferentMethods.java'))
        self.assertEqual(self.testClass.value(file), [28])

    def test_for_loops_in_nested_class(self):
        file = str(Path(self.cur_file_dir, 'NestedForInNestedClasses.java'))
        self.assertEqual(self.testClass.value(file), [9])

    def test_for_loops_in_anonymous_class(self):
        file = str(Path(self.cur_file_dir, 'ForInAnonymousFile.java'))
        self.assertEqual(self.testClass.value(file), [19])

    def test_nested_no_nested_if(self):
        pattern = NestedBlocks(2, BlockType.IF)
        file = str(Path(self.cur_file_dir, 'NestedNoIF.java'))
        self.assertEqual(pattern.value(file), [])

    def test_nested_if(self):
        pattern = NestedBlocks(2, BlockType.IF)
        file = str(Path(self.cur_file_dir, 'NestedIF.java'))
        self.assertEqual(pattern.value(file), [21, 42])



import os
from unittest import TestCase
from pathlib import Path
from aibolit.utils.lines import Lines


class TestLines(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_canReadLines(self):
        self.assertEqual(
            'class SimpleLinesTest {\n',
            (Lines(Path(self.dir_path, 'SimpleLinesTest.java')).value()[1])[0],
            'Did not match first line'
        )
        self.assertEqual(
            '    public void methodOne() {\n',
            (Lines(Path(self.dir_path, 'SimpleLinesTest.java')).value()[1])[1],
            'Did not match second line'
        )
        self.assertEqual(
            '        //do anything\n',
            (Lines(Path(self.dir_path, 'SimpleLinesTest.java')).value()[1])[2],
            'Did not match third line'
        )
        self.assertEqual(
            '    }\n',
            (Lines(Path(self.dir_path, 'SimpleLinesTest.java')).value()[1])[3],
            'Did not match fourth line'
        )
        self.assertEqual(
            '}',
            (Lines(Path(self.dir_path, 'SimpleLinesTest.java')).value()[1])[4],
            'Did not match last line'
        )



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.empty_rethrow.empty_rethrow import EmptyRethrow
from pathlib import Path


class TestEmptyRethrow(TestCase):
    cur_file_dir = Path(os.path.realpath(__file__)).parent
    testClass = EmptyRethrow()

    def test_empty(self):
        file = str(Path(self.cur_file_dir, 'Empty.java'))
        self.assertEqual(self.testClass.value(file), [])

    def test_anonymous(self):
        file = str(Path(self.cur_file_dir, 'Anonymous.java'))
        self.assertEqual(self.testClass.value(file), [19, 25])

    def test_both_catches(self):
        file = str(Path(self.cur_file_dir, 'BothCatches.java'))
        self.assertEqual(self.testClass.value(file), [6, 8])

    def test_instance_different_methods(self):
        file = str(Path(self.cur_file_dir, 'MultipleCatch.java'))
        self.assertEqual(self.testClass.value(file), [6])

    def test_sequential_catch(self):
        file = str(Path(self.cur_file_dir, 'SequentialCatch.java'))
        self.assertEqual(self.testClass.value(file), [6, 8])

    def test_sequential_catch_try(self):
        file = str(Path(self.cur_file_dir, 'SequentialCatchTry.java'))
        self.assertEqual(self.testClass.value(file), [6, 8, 13, 15])

    def test_simple(self):
        file = str(Path(self.cur_file_dir, 'Simple.java'))
        self.assertEqual(self.testClass.value(file), [6])

    def test_tricky_fake(self):
        file = str(Path(self.cur_file_dir, 'TrickyFake.java'))
        self.assertEqual(self.testClass.value(file), [])

    def test_try_inside_catch(self):
        file = str(Path(self.cur_file_dir, 'TryInsideCatch.java'))
        self.assertEqual(self.testClass.value(file), [10])

    def test_try_inside_finally(self):
        file = str(Path(self.cur_file_dir, 'TryInsideFinally.java'))
        self.assertEqual(self.testClass.value(file), [6, 12])

    def test_try_inside_try(self):
        file = str(Path(self.cur_file_dir, 'TryInsideTry.java'))
        self.assertEqual(self.testClass.value(file), [10, 17])

    def test_catch_with_functions(self):
        file = str(Path(self.cur_file_dir, 'CatchWithFunctions.java'))
        self.assertEqual(self.testClass.value(file), [13])

    def test_catch_with_if(self):
        file = str(Path(self.cur_file_dir, 'CatchWithIf.java'))
        self.assertEqual(self.testClass.value(file), [14])

    def test_chained_exceptions(self):
        file = str(Path(self.cur_file_dir, 'DataBaseNavigator.java'))
        file2 = str(Path(self.cur_file_dir, 'ConfigImportWizardPageDbvis.java'))
        file3 = str(Path(self.cur_file_dir, 'DatabaseNavigatorSourceContainer.java'))
        self.assertEqual(self.testClass.value(file), [])
        self.assertEqual(self.testClass.value(file2), [])
        self.assertEqual(self.testClass.value(file3), [])

    def test_try_without_catch(self):
        file = str(Path(self.cur_file_dir, 'ConcurrentDiskUtil.java'))
        self.assertEqual(self.testClass.value(file), [])

    def test_throw_with_cast(self):
        file = str(Path(self.cur_file_dir, 'CommonRdbmsReader.java'))
        self.assertEqual(self.testClass.value(file), [])



from aibolit.ast_framework import ASTNodeType
from aibolit.ast_framework.java_package import JavaPackage
from typing import Set, Any


class RFC:
    '''
    The Response For a Class is an object-oriented
    metric that shows the interaction of the class'
    methods with other methods.
    '''
    def __init__(self):
        self.class_methods = {}

    def exclude_inhereted_methods(self) -> Set[Any]:
        temp = self.class_methods.copy()
        for each_method in self.class_methods.keys():
            counter = 0
            for i in self.class_methods[each_method]:
                if i in self.class_methods:
                    counter += 1
                else:
                    counter -= 1
            if counter > 0:
                temp.pop(each_method)
        final_rfc = set()
        for declared_method in temp.keys():
            if len(declared_method) != 0:
                final_rfc.add(declared_method)

        for invoked_methods in temp.values():
            for each_method in invoked_methods:
                final_rfc.add(each_method)
        return final_rfc

    def get_invoked(self, tree) -> Set[Any]:
        inv_names = set()
        inv_methods = tree.get_nodes(ASTNodeType.METHOD_INVOCATION)
        for inv_method in inv_methods:
            name_of_invoked_class = tree.get_method_invocation_params(inv_method)
            current_name = name_of_invoked_class.method_name
            inv_names.add(current_name)
        return inv_names

    def value(self, filename: str) -> int:  # noqa: C901
        p = JavaPackage(filename)
        for class_name in p.java_classes:
            tree = p.java_classes[class_name]
            declareted_methods = tree.get_subtrees(ASTNodeType.METHOD_DECLARATION)
            for class_method in declareted_methods:
                # to form a set of all methods in the class
                names = list(class_method.children_with_type(class_method.root, ASTNodeType.STRING))
                for each_string in names:
                    method_name = tree.get_attr(each_string, 'string')
                    # we need to check the name because even comments are counted as the childs with string type
                    # need to get rid of them
                    if not method_name.startswith('/'):
                        self.class_methods[method_name] = set()
                        break

            # to count invoked methods
            tree = p.java_classes[class_name]
            declareted_methods = tree.get_subtrees(ASTNodeType.METHOD_DECLARATION)
            for meth_name, class_method in zip(self.class_methods.keys(), declareted_methods):
                invoked_names = self.get_invoked(class_method)
                for i in invoked_names:
                    self.class_methods[meth_name].add(i)

        final_rfc = len(self.exclude_inhereted_methods())
        self.class_methods.clear()
        return final_rfc



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import networkx as nx  # type: ignore
from networkx import Graph
from aibolit.utils.filter import Filters

from typing import List, Tuple, Union, TypeVar
from javalang.tree import ClassDeclaration, MethodDeclaration, \
    MemberReference, FieldDeclaration, MethodInvocation, This, Node, LocalVariableDeclaration


FldExh = Tuple[str, Tuple[str, str]]
MthExh = Tuple[str, Tuple[Tuple[str, str], ...]]
HasMember = Union[MemberReference, MethodInvocation]
HasSelector = Union[MemberReference, MethodInvocation, This]
T = TypeVar('T', bound=Node)
S = TypeVar('S', HasSelector, HasMember)
ThisNodes = Tuple[tuple, This]
SelMemNodes = Tuple[tuple, S]
RefNodes = Tuple[tuple, MemberReference]
InvNodes = Tuple[tuple, MethodInvocation]
LocalNodes = Tuple[tuple, LocalVariableDeclaration]
MthNodes = Tuple[tuple, MethodDeclaration]
Nodes = Tuple[tuple, T]
EdgeNode = Union[MthExh, FldExh]


class CohesionGraph:

    filtrate = Filters()

    def value(self, tree: Node) -> Graph:

        G: Graph = nx.Graph()

        for class_path, class_node in tree.filter(ClassDeclaration):

            # Extract all methods and fields
            # from ClassDeclaration node

            field_nodes: List[Nodes] = \
                self.filtrate.filter_node_lvl(class_node, FieldDeclaration)
            full_field_exhaust: List[FldExh] = \
                list(self.filtrate.exhaust_field(field_node) for path, field_node in field_nodes)
            clear_field_exhaust: List[FldExh] = \
                self.filtrate.clean_for_repetitions(full_field_exhaust)
            method_nodes: List[MthNodes] = \
                self.filtrate.filter_node_lvl(class_node, MethodDeclaration)
            method_nodes_filtered: List[MthNodes] = \
                self.filtrate.filter_getters_setters(method_nodes)
            full_method_exhaust: List[MthExh] = \
                list(self.filtrate.exhaust_method(method_node) for path, method_node in method_nodes_filtered)
            clear_method_exhaust: List[MthExh] = \
                self.filtrate.clean_for_repetitions(full_method_exhaust)

            for method in clear_method_exhaust:     # Add all MethodDeclarations to graph
                G.add_node(method[0] + str(hash(method[1])))
            for field in clear_field_exhaust:       # Add all FieldDeclarations to graph
                G.add_node(field[0] + str(hash(field[1])))

            for method_path, method_node in method_nodes_filtered:
                # Find and compare all existing
                # MemberReferences, MethodInvocations,
                # This statements, LocalVariableDeclarations
                # to themselves and objects added to graph G
                reference_nodes: List[RefNodes] = \
                    self.filtrate.filter_node_lvl(method_node, MemberReference)
                invocation_nodes: List[InvNodes] = \
                    self.filtrate.filter_node_lvl(method_node, MethodInvocation)
                this_nodes: List[ThisNodes] = \
                    self.filtrate.filter_node_lvl(method_node, This)
                local_nodes: List[LocalNodes] = \
                    self.filtrate.filter_node_lvl(method_node, LocalVariableDeclaration)
                local_exhaust: List[FldExh] = \
                    list(self.filtrate.exhaust_field(local_node) for path, local_node in local_nodes)
                method_exhaust: MthExh = self.filtrate.exhaust_method(method_node)

                self.add_references_to_graph(G, reference_nodes, local_exhaust, clear_field_exhaust, method_exhaust)
                self.add_this_to_graph(G, this_nodes, clear_field_exhaust, method_exhaust)
                self.add_invocations_to_graph(
                    G, invocation_nodes, clear_method_exhaust, method_exhaust, local_exhaust, clear_field_exhaust)
            break  # Stop after first class
        return G

    # ------------------------------------------------
    # Funcs for adding edges to graph

    def add_invocations_to_graph(
        self,
        G: Graph,
        invocation_nodes: List[SelMemNodes],
        full_method_exhaust: List[MthExh],
        method_exhaust: MthExh,
        local_exhaust: List[FldExh],
        full_field_exhaust: List[FldExh]
    ) -> None:
        """Adds nodes to graph G

        Gets list of invocation names as input and
        compares them to existing list of exhausted "MethodDeclarations"
        After successful comparison calls "add_vertices_edges"
        Adding nodes and edges between.
        """
        for invocation_path, invocation_node in invocation_nodes:
            if isinstance(invocation_node.selectors, list):  # Check for inv being first in whole statement
                for method in full_method_exhaust:
                    if invocation_node.member == method[0]:
                        self.add_vertices_edges(G, 'invocation', method_exhaust, method)
                inv_arguments = self.filtrate.get_arguments(invocation_node)
                inv_fields: List[str] = inv_arguments[1]
                inv_funcs: List[str] = inv_arguments[0]
                if len(inv_fields) > 0:
                    self.add_invocation_fields(G, inv_fields, local_exhaust, full_field_exhaust, method_exhaust)
                if len(inv_funcs) > 0:
                    self.add_invocation_funcs(G, inv_funcs, full_method_exhaust, method_exhaust)

    def add_invocation_fields(
        self,
        G: Graph,
        inv_fields: List[str],
        local_exhaust: List[FldExh],
        full_field_exhaust: List[FldExh],
        method_exhaust: MthExh
    ) -> None:
        """Adds nodes to graph G

        Gets a list of invocation field attributes as input and
        compares them to existing list of exhausted "FieldDeclarations"
        After successful comparison calls "add_vertices_edges"
        Adding nodes and edges between.
        """
        for inv_argument in inv_fields:
            if inv_argument not in [x[0] for x in local_exhaust]:
                for field in full_field_exhaust:
                    if inv_argument == field[0]:
                        self.add_vertices_edges(G, 'reference', method_exhaust, field)

    def add_invocation_funcs(
        self,
        G: Graph,
        inv_funcs: List[str],
        full_method_exhaust: List[MthExh],
        method_exhaust: MthExh
    ) -> None:
        """Adds nodes to graph G

        Gets a list of invocation method attributes as input and
        compares them to existing list of exhausted "MethodDeclarations"
        After successful comparison calls "add_vertices_edges"
        Adding nodes and edges between.
        """
        # ToDo: make a func for a return type check
        for inv_argument in inv_funcs:
            for method in full_method_exhaust:
                if inv_argument == method[0]:
                    self.add_vertices_edges(G, 'invocation', method_exhaust, method)

    def add_references_to_graph(
        self,
        G: Graph,
        reference_nodes: List[SelMemNodes],
        local_exhaust: List[FldExh],
        full_field_exhaust: List[FldExh],
        method_exhaust: MthExh
    ) -> None:
        """Adds nodes to graph G

        Gets list of "MemberReferences" nodes as input and
        compares them to existing list of exhausted FieldDeclarations.
        After successful comparison calls "add_vertices_edges"
        Adding nodes and edges between.
        """
        for reference_path, reference_node in reference_nodes:
            if isinstance(reference_node.selectors, list):  # Check for node being "alone"
                if reference_node.member not in [x[0] for x in local_exhaust] and \
                        reference_node.member not in [x for x, _ in method_exhaust[1]]:
                    for field in full_field_exhaust:
                        if reference_node.member == field[0]:
                            self.add_vertices_edges(G, 'reference', method_exhaust, field)

    def add_this_to_graph(
        self,
        G: Graph,
        this_nodes: List[ThisNodes],
        full_field_exhaust: List[FldExh],
        method_exhaust: MthExh
    ) -> None:
        """Adds nodes to graph G

        Gets list of "This" nodes as input and
        compares them to existing list of exhausted "FieldDeclarations"
        After successful comparison calls "add_vertices_edges"
        Adding nodes and edges between.
        """
        for this_path, this_node in this_nodes:
            for field in full_field_exhaust:
                if len(this_node.selectors) == 1 and isinstance(this_node.selectors[0], MemberReference):
                    if this_node.selectors[0].member in field:
                        self.add_vertices_edges(G, 'reference', method_exhaust, field)

    def add_vertices_edges(self, G, edge_type: str, first_node: EdgeNode, second_node: EdgeNode) -> None:
        """Adds nodes to graph G

        Gets two objects as input and
        adds nodes and an edge between.
        If nodes already exist:
        creates an edge between
        """

        G.add_node(first_node[0] + str(hash(first_node[1])))
        G.add_node((second_node[0]) + str(hash(second_node[1])))
        G.add_edge(first_node[0] + str(hash(first_node[1])),
                   (second_node[0]) + str(hash(second_node[1])), type=edge_type)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from unittest import TestCase

from javalang.parser import Parser
from javalang.tokenizer import tokenize

from aibolit.patterns.null_check.null_check import NullCheck


class TestNullCheck(TestCase):
    def test_null_check(self):
        snippet = """\
        if (this.z == null) { // here!
            throw new RuntimeException("oops");
        }
        """

        # @todo #116:30min `_traverse_node` is a private method, so should not
        #  be called directly. Requires change to `Pattern` interface. Should
        #  be done for each test in this class.
        self.assertEqual(
            NullCheck()._traverse_node(_parser(snippet).parse_block_statement()), [1]
        )

    def test_null_check_in_constructor(self):
        snippet = """\
        public NullCheck(String z) {
            if (z == null) { // here!
                throw new RuntimeException("oops");
            }
        }
        """

        self.assertEqual(
            NullCheck()._traverse_node(_parser(snippet).parse_member_declaration()),
            [],
        )

    def test_null_check_comparison_result_assignment(self):
        snippet = "boolean i = z == null;"

        self.assertEqual(
            NullCheck()._traverse_node(_parser(snippet).parse_block_statement()), [1]
        )

    def test_null_check_ternary(self):
        snippet = 'luckyName == null ? luckyName : "No lucky name found";'

        self.assertEqual(
            NullCheck()._traverse_node(_parser(snippet).parse_block_statement()), [1]
        )

    def test_null_check_not_equal_comparison(self):
        snippet = """\
        if (this.z != null) { // here!
            throw new RuntimeException("oops");
        }
        """

        self.assertEqual(
            NullCheck()._traverse_node(_parser(snippet).parse_block_statement()), [1]
        )

    def test_null_check_using_methods(self):
        # @todo #116:30min Implement `null` check for those patterns as well:
        #  assertThrows(NullPointerException.class, () -> wrapperSum(null, 2));
        #  Optional op2 = Optional.ofNullable(null);
        #  Objects.requireNonNull(bar, "bar must not be null");
        pass


def _parser(snippet):
    return Parser(tokenize(snippet))



import unittest
from aibolit.metrics.mda.mda import MDAMetric
from pathlib import Path
import os


class TestMDAMetric(unittest.TestCase):
    mda_metric = MDAMetric()
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_book(self):

        file = Path(self.dir_path, 'Book.java')
        result = self.mda_metric.value(file)
        self.assertEqual(result, 6)

    def test_one(self):

        file = Path(self.dir_path, 'FirstTest.java')
        result = self.mda_metric.value(file)
        self.assertEqual(result, 8)

    def test_two(self):

        file = Path(self.dir_path, 'SecondTest.java')
        result = self.mda_metric.value(file)
        self.assertEqual(result, 7)

    def test_three(self):

        file = Path(self.dir_path, 'ThirdTest.java')
        result = self.mda_metric.value(file)
        self.assertEqual(result, 9)



from aibolit.utils.ast_builder import build_ast


class MDAMetric:
    @staticmethod
    def get_depth(tree) -> int:
        result = 0
        max_val = 0

        if tree is None:
            return 0
        elif isinstance(tree, (list, set, tuple)):
            children = tree
        elif not hasattr(tree, "children"):
            return 0
        else:
            children = tree.children
            result = 1

        for child in children:
            cur_val = MDAMetric.get_depth(child)
            if cur_val > max_val:
                max_val = cur_val
        return max_val + result

    @staticmethod
    def value(filename: str) -> int:
        tree = build_ast(filename)
        metric = MDAMetric.get_depth(tree)

        return metric




import os
import unittest
from unittest import TestCase
from aibolit.patterns.non_final_argument.non_final_argument import NonFinalArgument
from pathlib import Path


@unittest.skip('Not implemented')
class TestNonFinalArgument(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_non_final_argument_in_ctor(self):
        self.assertEqual(
            NonFinalArgument().value(Path(self.dir_path, 'NonFinalArgumentCtor.java')),
            [7]
        )

    def test_find_non_final_argument_in_method(self):
        self.assertEqual(
            NonFinalArgument().value(Path(self.dir_path, 'NonFinalArgumentMethod.java')),
            [11]
        )



import os
from unittest import TestCase
from pathlib import Path
from aibolit.patterns.array_as_argument.array_as_argument import ArrayAsArgument


class TestArrayAsArgument(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    pattern = ArrayAsArgument()

    def test_NoArgument(self):
        file = Path(self.dir_path, 'NoArgument.java')
        self.assertEqual(
            [],
            self.pattern.value(file),
            'Should not match no argument method'
        )

    def test_PrimitiveAsArgument(self):
        file = Path(self.dir_path, 'PrimitiveAsArgument.java')
        self.assertEqual(
            [],
            self.pattern.value(file),
            'Should not match method with primitive as argument'
        )

    def test_ArrayAsArgument(self):
        file = Path(self.dir_path, 'ArrayAsArgument.java')
        self.assertEqual(
            [2],
            self.pattern.value(file),
            'Should match method with array as argument'
        )

    def test_ObjectAsArgument(self):
        file = Path(self.dir_path, 'ObjectAsArgument.java')
        self.assertEqual(
            [],
            self.pattern.value(file),
            'Should not match method with object as argument'
        )

    def test_PrimitiveAndArrayAsArgument(self):
        file = Path(self.dir_path, 'PrimitiveAndArrayAsArgument.java')
        self.assertEqual(
            [2],
            self.pattern.value(file),
            'Should match method with array as argument'
        )

    def test_GenericArrayAsArgument(self):
        file = Path(self.dir_path, 'GenericArrayAsArgument.java')
        self.assertEqual(
            [2],
            self.pattern.value(file),
            'Should match method with generic array as argument'
        )

    def test_ConstructorWithArrayAsArgument(self):
        file = Path(self.dir_path, 'ConstructorWithArrayAsArgument.java')
        self.assertEqual(
            [2],
            self.pattern.value(file),
            'Should match constructor with array as argument'
        )



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
from unittest import TestCase
from aibolit.patterns.method_chaining.method_chaining import MethodChainFind
from pathlib import Path


class TestMethodChain(TestCase):
    dir_path = Path(os.path.realpath(__file__)).parent
    method_chain_finder = MethodChainFind()

    def test_method_chain(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChain.java'))
        self.assertEqual(len(lines), 1)

    def test_empty_method_chain(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'EmptyMethodChain.java'))
        self.assertEqual(len(lines), 1)

    def test_chain_with_new_object(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChainNewObjectMethods.java'))
        self.assertEqual(len(lines), 1)

    def test_method_chain_in_different_methods(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChainInDifferentMethods.java'))
        self.assertEqual(len(lines), 2)

    def test_chain_in_nested_class(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChainNestedClass.java'))
        self.assertEqual(len(lines), 1)

    def test_chain_in_anonymous_class(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChainAnonymousClass.java'))
        self.assertEqual(len(lines), 1)

    def test_chain_in_anonymous_class_empty(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChainAnonymousClassEmpty.java'))
        self.assertEqual(len(lines), 0)

    def test_several_chains(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MethodChainSeveral.java'))
        self.assertEqual(len(lines), 3)

    def test_chain_without_object_creating(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'WithoutObjectCreating.java'))
        self.assertEqual(len(lines), 1)

    def test_nested_chain_with_this(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'NestedChainWIthThis.java'))
        self.assertEqual(len(lines), 2)

    def test_nested_chain_with_simple_method_invocation(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'NestedChainWithSimpleMethodInvocation.java'))
        self.assertEqual(len(lines), 2)

    def test_nested_chain_complicated_structure(self):
        """
        Several nested structures are checked: nested method chaining
        with nested anonymous classes
        """
        lines = self.method_chain_finder.value(Path(self.dir_path, 'HolyMolyNestedChain.java'))
        self.assertEqual(len(lines), 3)

    def test_smallest_chain(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'SmallestChain.java'))
        self.assertEqual(len(lines), 1)

    def test_fake_chain(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'FakeChain.java'))
        self.assertEqual(len(lines), 0)

    def test_many_chains(self):
        lines = self.method_chain_finder.value(Path(self.dir_path, 'MachineLearningGetResultsIT.java'))
        self.assertGreater(len(lines), 300)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from collections import defaultdict

import javalang

from aibolit.utils.java_parser import JavalangImproved


class PartialSync:

    def __init__(self):
        pass

    def value(self, filename):
        total_code_lines = set()
        obj = JavalangImproved(filename)
        empty_lines = obj.get_empty_lines()
        items = obj.tree_to_nodes()
        synch_nodes = defaultdict(list)
        method_nodes = {}

        for x in items:
            is_instance_sync_stat = isinstance(x.node, javalang.tree.SynchronizedStatement)
            is_instance_method_decl = isinstance(x.node, javalang.tree.MethodDeclaration)
            is_instance_ctr = isinstance(x.node, javalang.tree.ConstructorDeclaration)
            is_instance_lambda = isinstance(x.node, javalang.tree.LambdaExpression)

            if is_instance_sync_stat:
                synch_nodes[x.method_line].append(x)
            elif is_instance_method_decl or is_instance_ctr or is_instance_lambda:
                method_nodes[x.method_line] = x

        for method_line, sync_nodes in sorted(synch_nodes.items(), key=lambda x: x[1][0].line):
            for sync_n in sync_nodes:
                lines = set(range(method_line, sync_n.line))
                empty_lines_before_sync = [x for x in lines if x in empty_lines]
                lines_number_btw_function_and_synch_block = sync_n.line - method_line - len(
                    empty_lines_before_sync)
                if lines_number_btw_function_and_synch_block > 1:
                    total_code_lines.add(sync_n.line)
                    continue
                elif lines_number_btw_function_and_synch_block == 1:
                    # if there is any statement after SynchronizedStatement, then it's an anti-pattern
                    method_item = method_nodes[method_line]
                    if len(method_item.node.body) > 1 and \
                            isinstance(method_item.node.body[0], javalang.tree.SynchronizedStatement):
                        total_code_lines.add(sync_n.line)
                        continue

        return sorted(total_code_lines)



# The MIT License (MIT)
#
# Copyright (c) 2020 Aibolit
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included
# in all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NON-INFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import os
import unittest
from pathlib import Path
from unittest import TestCase
from aibolit.patterns.loop_outsider.loop_outsider import LoopOutsider


@unittest.skip('Not implemented')
class TestLoopOutsider(TestCase):
    """
    @todo #138:30min Continue to implement loop_outsider
     Implement loop_outsider tests regarding for and do-while loops. Then implement
     loop_outsider pattern itself.
    """
    dir_path = Path(os.path.realpath(__file__)).parent

    def test_find_loop_outsider_assignment_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderAssignmentInWhile.java')),
            [5],
            'Could not find loop outsider assignment in while loop'
        )

    def test_find_loop_outsider_prefix_increment_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderPrefixIncrementInWhile.java')),
            [5],
            'Could not find loop outsider with prefix increment in while loop'
        )

    def test_find_loop_outsider_postfix_increment_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderPostfixIncrementInWhile.java')),
            [5],
            'Could not find loop outsider with postfix increment in while loop'
        )

    def test_find_loop_outsider_prefix_decrement_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderPrefixDecrementInWhile.java')),
            [5],
            'Could not find loop outsider with prefix decrement in while loop'
        )

    def test_find_loop_outsider_postfix_decrement_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderPostfixDecrementInWhile.java')),
            [5],
            'Could not find loop outsider with postfix decrement in while loop'
        )

    def test_find_loop_outsider_add_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderAddAndInWhile.java')),
            [5],
            'Could not find loop outsider with add AND in while loop'
        )

    def test_find_loop_outsider_subtract_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderSubtractAndInWhile.java')),
            [5],
            'Could not find loop outsider with subtract AND in while loop'
        )

    def test_find_loop_outsider_multiply_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderMultiplyAndInWhile.java')),
            [5],
            'Could not find loop outsider with multiply AND in while loop'
        )

    def test_find_loop_outsider_divide_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderDivideAndInWhile.java')),
            [5],
            'Could not find loop outsider with divide AND in while loop'
        )

    def test_find_loop_outsider_modulus_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderModulusAndInWhile.java')),
            [5],
            'Could not find loop outsider with modulus AND in while loop'
        )

    def test_find_loop_outsider_left_shift_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderLeftShiftAndInWhile.java')),
            [5],
            'Could not find loop outsider with left shift AND in while loop'
        )

    def test_find_loop_outsider_right_shift_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderRightShiftAndInWhile.java')),
            [5],
            'Could not find loop outsider with right shift AND in while loop'
        )

    def test_find_loop_outsider_bitwise_and_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderBitwiseAndInWhile.java')),
            [5],
            'Could not find loop outsider with bitwise AND in while loop'
        )

    def test_find_loop_outsider_bitwise_inclusive_or_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderBitwiseInclusiveOrInWhile.java')),
            [5],
            'Could not find loop outsider with bitwise inclusive OR in while loop'
        )

    def test_find_loop_outsider_bitwise_exclusive_or_in_while(self):
        self.assertEqual(
            LoopOutsider().value(Path(self.dir_path, 'LoopOutsiderBitwiseExclusiveOrInWhile.java')),
            [5],
            'Could not find loop outsider with bitwise exclusive OR in while loop'
        )
